@@ -5,8 +5,10 @@
 #include <vector>
 #include <string>
 #include <cstring>
+#include <limits>
 extern "C" {
 #include "lua.h"
+#include "luajit/src/lj_def.h"
 }
 #define R_NO_REMAP
 #include <R.h>
@@ -21,7 +23,7 @@ template <typename Push>
 static void push_R_vector(lua_State* L, SEXP x, char as, int type, Push push)
 {
     // Get length of vector
-    int len = Rf_length(x);
+    R_xlen_t xlen = Rf_xlength(x);
 
     switch (as)
     {
@@ -32,7 +34,7 @@ static void push_R_vector(lua_State* L, SEXP x, char as, int type, Push push)
             // Call it with arguments x as userdata, type code as integer
             lua_pushlightuserdata(L, x);
             lua_pushinteger(L, type | REFERENCE_T);
-            luajr_pcall(L, 2, 1, "luajr.construct_ref() from push_R_vector()", LUAJR_TOOLING_ALL);
+            luajr_pcall(L, 2, 1, "luajr.construct_ref() from push_R_vector()", LUAJR_TOOLING_NONE);
             break;
 
         case 'v':
@@ -42,32 +44,35 @@ static void push_R_vector(lua_State* L, SEXP x, char as, int type, Push push)
             // Call it with arguments x as userdata, type code as integer
             lua_pushlightuserdata(L, x);
             lua_pushinteger(L, type | VECTOR_T);
-            luajr_pcall(L, 2, 1, "luajr.construct_vec() from push_R_vector()", LUAJR_TOOLING_ALL);
+            luajr_pcall(L, 2, 1, "luajr.construct_vec() from push_R_vector()", LUAJR_TOOLING_NONE);
             break;
 
         case 's':
         case 'a':
-            if (len == 0)
+            if (xlen == 0)
                 lua_pushnil(L); // Length 0: push nil
-            else if (len == 1 && as == 's')
+            else if (xlen == 1 && as == 's')
                 push(L, x, 0);  // Length 1 and 's': push scalar
-            else
+            else if (xlen < LJ_MAX_ASIZE) // Strict < needed here.
             {                   // Length >1 or 'a': push table
-                lua_createtable(L, len, 0);
-                for (int i = 0; i < len; ++i)
+                lua_createtable(L, xlen, 0);
+                for (R_xlen_t i = 0; i < xlen; ++i)
                 {
                     push(L, x, i);
                     lua_rawseti(L, -2, i + 1);
                 }
             }
+            else
+                Rf_error("Cannot create Lua table with more than %d elements. Requested size: %.0f. Use 'r' or 'v' arg code instead.",
+                    LJ_MAX_ASIZE - 1, (double)xlen);
             break;
 
         default:
             if (as >= '1' && as <= '9')
             {
                 int reqn = as - '0';
-                if (len != reqn)
-                    Rf_error("Vector of length %d requested, but passed vector of length %d.", reqn, len);
+                if (xlen != reqn)
+                    Rf_error("Vector of length %d requested, but passed vector of length %.0f.", reqn, (double)xlen);
                 push_R_vector(L, x, 's', type, push);
                 break;
             }
@@ -80,6 +85,10 @@ static void push_R_vector(lua_State* L, SEXP x, char as, int type, Push push)
 static void push_R_list(lua_State* L, SEXP x, char as)
 {
     // Get length of vector
+    R_xlen_t xlen = Rf_xlength(x);
+    if (xlen >= LJ_MAX_ASIZE || xlen >= std::numeric_limits<int>::max())
+        Rf_error("List is too large to be passed to Lua. Cannot create Lua table with more than %d elements. Requested size: %.0f.",
+            LJ_MAX_ASIZE - 1, (double)xlen);
     int len = Rf_length(x);
 
     // Get names of list elements
@@ -135,7 +144,7 @@ static void push_R_list(lua_State* L, SEXP x, char as)
             }
 
             // Call luajr.construct_list
-            luajr_pcall(L, 2, 1, "luajr.construct_list() from push_R_list()", LUAJR_TOOLING_ALL);
+            luajr_pcall(L, 2, 1, "luajr.construct_list() from push_R_list()", LUAJR_TOOLING_NONE);
 
             break;
 
@@ -172,6 +181,14 @@ static void push_R_list(lua_State* L, SEXP x, char as)
     UNPROTECT(1);
 }
 
+// Stop if length of chr exceeds LuaJIT limits.
+void CheckStringLength(SEXP chr)
+{
+    R_xlen_t xlen = Rf_xlength(chr);
+    if (xlen >= LJ_MAX_STR)
+        Rf_error("Cannot pass string with more than %d bytes. Requested size: %.0f.", LJ_MAX_STR, (double)xlen);
+}
+
 // Analogous to Lua's lua_pushXXX(lua_State* L, XXX x) functions, this pushes
 // the R object [x] onto Lua's stack.
 // Supported: NILSXP, LGLSXP, INTSXP, REALSXP, STRSXP, VECSXP, EXTPTRSXP,
@@ -188,7 +205,7 @@ extern "C" void luajr_pushsexp(lua_State* L, SEXP x, char as)
             {
                 lua_pushlightuserdata(L, (void*)&luajr_construct_null);
                 lua_rawget(L, LUA_REGISTRYINDEX);
-                luajr_pcall(L, 0, 1, "luajr.construct_null() from luajr_pushsexp()", LUAJR_TOOLING_ALL);
+                luajr_pcall(L, 0, 1, "luajr.construct_null() from luajr_pushsexp()", LUAJR_TOOLING_NONE);
             }
             else
                 lua_pushnil(L);
@@ -211,7 +228,8 @@ extern "C" void luajr_pushsexp(lua_State* L, SEXP x, char as)
         case STRSXP: // character vector: r, v, s, a, 1-9
             push_R_vector(L, x, as, CHARACTER_T,
                 [](lua_State* L, SEXP x, unsigned int i)
-                    { lua_pushstring(L, CHAR(STRING_ELT(x, i))); });
+                    { CheckStringLength(STRING_ELT(x, i));
+                      lua_pushstring(L, CHAR(STRING_ELT(x, i))); });
             break;
         case VECSXP: // list (generic vector): r, v, s
             push_R_list(L, x, as);
@@ -220,6 +238,7 @@ extern "C" void luajr_pushsexp(lua_State* L, SEXP x, char as)
             lua_pushlightuserdata(L, R_ExternalPtrAddr(x));
             break;
         case RAWSXP: // raw bytes
+            CheckStringLength(x);
             lua_pushlstring(L, (const char*)RAW(x), Rf_length(x));
             break;
         default:
@@ -275,7 +294,7 @@ extern "C" SEXP luajr_tosexp(lua_State* L, int index)
 
             // Call it with table arg
             lua_pushvalue(L, index);
-            luajr_pcall(L, 1, 2, "luajr.return_info() from luajr_tosexp() [1]", LUAJR_TOOLING_ALL);
+            luajr_pcall(L, 1, 2, "luajr.return_info() from luajr_tosexp() [1]", LUAJR_TOOLING_NONE);
 
             // If not a known table type, return normal table
             if (lua_isnil(L, -2))
@@ -284,16 +303,25 @@ extern "C" SEXP luajr_tosexp(lua_State* L, int index)
                 lua_pop(L, 2); // clear two return values
 
                 // First pass: count table entries of each type.
-                int narr = 0, nrec = 0;
+                size_t narr = 0, nrec = 0;
                 lua_pushnil(L);
                 while (lua_next(L, index) != 0)
                 {
                     if      (lua_type(L, -2) == LUA_TNUMBER) ++narr;
                     else if (lua_type(L, -2) == LUA_TSTRING) ++nrec;
-                    else    Rf_error("Lua type %s keys cannot be represented in an R list.", lua_typename(L, lua_type(L, -2)));
+                    else
+                    {
+                        lua_pop(L, 2);
+                        Rf_error("Lua type %s keys cannot be represented in an R list.",
+                            lua_typename(L, lua_type(L, -2)));
+                    }
                     lua_pop(L, 1);
                 }
 
+                if (narr + nrec >= LJ_MAX_ASIZE)
+                    Rf_error("Table is too large to be returned to R. Limit is %d elements. Requested size: %.0f.",
+                        LJ_MAX_ASIZE - 1, (double)(narr + nrec));
+
                 // Create list and names
                 SEXP retval = PROTECT(Rf_allocVector(VECSXP, narr + nrec));
                 SEXP names = R_NilValue;
@@ -318,11 +346,12 @@ extern "C" SEXP luajr_tosexp(lua_State* L, int index)
                         ++rec_i;
                     }
                     lua_pop(L, 1);
+                    UNPROTECT(1);
                 }
 
                 // Name and return
                 Rf_setAttrib(retval, R_NamesSymbol, names);
-                UNPROTECT(1 + (nrec > 0) + narr + nrec);
+                UNPROTECT(1 + (nrec > 0));
                 return retval;
             }
 
@@ -334,9 +363,12 @@ extern "C" SEXP luajr_tosexp(lua_State* L, int index)
             // List type
             if (type == LIST_T)
             {
+                if (size >= LJ_MAX_ASIZE)
+                    Rf_error("List is too large to be returned to R. Limit is %d elements. Requested size: %.0f.",
+                        LJ_MAX_ASIZE - 1, (double)size);
+
                 // Add each entry to a list
                 SEXP retval = PROTECT(Rf_allocVector(VECSXP, size));
-                int nprotect = 1;
 
                 // Get the list contents on the stack
                 lua_rawgeti(L, index, 0); // get list[0]
@@ -346,7 +378,6 @@ extern "C" SEXP luajr_tosexp(lua_State* L, int index)
                 while (lua_next(L, -2) != 0)
                 {
                     SEXP val = PROTECT(luajr_tosexp(L, -1));
-                    ++nprotect;
                     if (lua_type(L, -2) == LUA_TNUMBER) // List element
                     {
                         SET_VECTOR_ELT(retval, lua_tointeger(L, -2) - 1, val);
@@ -359,7 +390,6 @@ extern "C" SEXP luajr_tosexp(lua_State* L, int index)
                             if (size > 0)
                             {
                                 SEXP names = PROTECT(Rf_allocVector(STRSXP, size));
-                                ++nprotect;
                                 SEXP setnames = Rf_getAttrib(val, R_NamesSymbol);
                                 for (int i = 0; i < Rf_length(val); ++i) {
                                     int index = REAL(VECTOR_ELT(val, i))[0] - 1;
@@ -367,6 +397,7 @@ extern "C" SEXP luajr_tosexp(lua_State* L, int index)
                                     SET_STRING_ELT(names, index, name);
                                 }
                                 Rf_setAttrib(retval, R_NamesSymbol, names);
+                                UNPROTECT(1);
                             }
                         }
                         else
@@ -379,27 +410,28 @@ extern "C" SEXP luajr_tosexp(lua_State* L, int index)
                         Rf_error("Lua type %s keys cannot be represented in an R list.",
                             lua_typename(L, lua_type(L, -2)));
                     }
+                    UNPROTECT(1);
                     lua_pop(L, 1);
                 }
 
                 // Special check: ensure data.frame has rownames attribute
                 // If retval is data.frame with at least one column, but no row names:
-                if (Rf_inherits(retval, "data.frame") && Rf_length(retval) > 0 &&
+                if (Rf_inherits(retval, "data.frame") && size > 0 &&
                     Rf_getAttrib(retval, R_RowNamesSymbol) == R_NilValue)
                 {
                     // R has a special shorthand for "short" rownames: c(NA_integer_, nrow) (see attrib.c)
                     // TODO This will fail if the number of rows is 2^31 or higher. However, it also seems
-                    // that R itself cannot really handle such long data.frames (at least as of R 4.3.2).
+                    // that R itself cannot handle such long data.frames either (as of R 4.5.1).
                     SEXP rownames = PROTECT(Rf_allocVector(INTSXP, 2));
-                    ++nprotect;
                     INTEGER(rownames)[0] = NA_INTEGER;
                     INTEGER(rownames)[1] = Rf_length(VECTOR_ELT(retval, 0));
                     Rf_setAttrib(retval, R_RowNamesSymbol, rownames);
+                    UNPROTECT(1);
                 }
 
                 // Return
                 lua_pop(L, 1); // pop list[0]
-                UNPROTECT(nprotect);
+                UNPROTECT(1);
                 return retval;
             }
 
@@ -417,7 +449,7 @@ extern "C" SEXP luajr_tosexp(lua_State* L, int index)
             // Call it with cdata arg and pointer
             lua_pushvalue(L, index);
             lua_pushlightuserdata(L, ret);
-            luajr_pcall(L, 2, 0, "luajr.return_copy() from luajr_tosexp() [1]", LUAJR_TOOLING_ALL);
+            luajr_pcall(L, 2, 0, "luajr.return_copy() from luajr_tosexp() [1]", LUAJR_TOOLING_NONE);
             // Return SEXP
             UNPROTECT(1);
             return ret;
@@ -443,7 +475,7 @@ extern "C" SEXP luajr_tosexp(lua_State* L, int index)
             lua_rawget(L, LUA_REGISTRYINDEX);
             // Call it with cdata arg
             lua_pushvalue(L, index);
-            luajr_pcall(L, 1, 2, "luajr.return_info() from luajr_tosexp() [2]", LUAJR_TOOLING_ALL);
+            luajr_pcall(L, 1, 2, "luajr.return_info() from luajr_tosexp() [2]", LUAJR_TOOLING_NONE);
 
             // If not a known cdata type, return external pointer
             if (lua_isnil(L, -2))
@@ -463,7 +495,7 @@ extern "C" SEXP luajr_tosexp(lua_State* L, int index)
                 // Call it with cdata arg and sexp
                 lua_pushvalue(L, index);
                 lua_pushlightuserdata(L, &ret);
-                luajr_pcall(L, 2, 0, "luajr.return_copy() from luajr_tosexp() [2]", LUAJR_TOOLING_ALL);
+                luajr_pcall(L, 2, 0, "luajr.return_copy() from luajr_tosexp() [2]", LUAJR_TOOLING_NONE);
                 // Return SEXP
                 return ret;
             }
@@ -494,7 +526,7 @@ extern "C" SEXP luajr_tosexp(lua_State* L, int index)
                 else if (rtype == INTSXP)   lua_pushlightuserdata(L, INTEGER(ret));
                 else if (rtype == REALSXP)  lua_pushlightuserdata(L, REAL(ret));
                 else Rf_error("Unknown type");
-                luajr_pcall(L, 2, 0, "luajr.return_copy() from luajr_tosexp() [3]", LUAJR_TOOLING_ALL);
+                luajr_pcall(L, 2, 0, "luajr.return_copy() from luajr_tosexp() [3]", LUAJR_TOOLING_NONE);
                 // Return SEXP
                 UNPROTECT(1);
                 return ret;
